/*
    Alternative Timer Message Example

    This example is identical to "MessageTimers.c" except that
    it uses the function AvrXCancelTimerMessage() to remove the
    timer from the time queue manager and restart it.
    
*/
#include <io.h>                 // Avr-gcc io definitions/macros
#include <signal.h>
#include "avrx.h"               // AvrX System calls/data structures
#include "hardware.h"           // Sample code hardware defines

void InitSerialIO(uint8_t);     // From Avrx.a (debug monitor)

// Declare all my AvrX data structures

AVRX_TASK(Monitor, 20, 0);      // External task definition

TimerControlBlock SwTimer;      // Plain vanilla timer
TimerMessageBlock Timer;        // TimerMessage timer (just a timer with a
                                // message queue pointer appended)

// This is how you build up a message element that transports data

typedef struct MyMessage
{
    MessageControlBlock mcb;
    uint8_t data;
}
*pMyMessage, MyMessage;

MyMessage SwitchMsg;            // Declare the actual message
MessageQueue MyQueue;           // The message queue itself


#define TCNT0_INIT (0xFF-CPUCLK/256/TICKRATE)
#define TMC8_CK256 (1<<CS02)
#define UBRR_INIT (CPUCLK/(16*BAUDRATE)-1)
/*
 Timer 0 Overflow Interrupt Handler

 Prototypical Interrupt handler:
 . Switch to kernel context
 . handle interrupt
 . switch back to interrupted context.
 */
AVRX_SIGINT(SIG_OVERFLOW0)
{
    IntProlog();                // Save interrupted context, switch stacks
    outp(TCNT0_INIT, TCNT0);    // Reload the timer counter
    AvrXTimerHandler();         // Process Timer queue
    Epilog();                   // Restore context of next running task
}

AVRX_TASKDEF(flasher, 20, 2)
{
    unsigned char led = 0;
    pMyMessage pMsg;

    outp(-1, LED-1);            // Initialize LED port to outputs
    
    while(1)
    {
        
        AvrXStartTimerMessage(&Timer, 150, &MyQueue);
        outp(~led, PORTB);
        pMsg = (pMyMessage)AvrXWaitMessage(&MyQueue);
        if (pMsg == &SwitchMsg)
        {
            led ^= ~(pMsg->data);
            AvrXAckMessage(&pMsg->mcb);
            if (AvrXCancelTimerMessage(&Timer, &MyQueue) != &Timer.u.mcb)
                AvrXHalt();     // The timer should always be found...
        }
        else if ((pTimerMessageBlock)pMsg == &Timer)
        {
            asm("rol %0\n"\
                "\tbrcc .+2\n"\
                "\tinc %0\n"  ::"r" (led));
        }
        else
        {
            AvrXHalt();
        }
    }
}
AVRX_TASKDEF(switcher, 20, 2)
{
    outp(-1, SWITCH+2);             // Enable pullups on switch inputs
    while(1)
    {
        AvrXDelay(&SwTimer, 10);    // Delay 10 milliseconds
        if (SwitchMsg.data != inp(SWITCH)) // On change, send message
        {
            SwitchMsg.data = inp(SWITCH);
            AvrXSendMessage(&MyQueue, &SwitchMsg.mcb);
            AvrXWaitMessageAck(&SwitchMsg.mcb);
        }
    }
}

int main(void)
{
    outp((1<<SE) , MCUCR);      // Enable "sleep" mode (low power when idle)
    outp(TCNT0_INIT, TCNT0);    // Load overflow counter of timer0
    outp(TMC8_CK256 , TCCR0);   // Set Timer0 to CPUCLK/256
    outp((1<<TOIE0), TIMSK);    // Enable interrupt flag

    AvrXRunTask(TCB(flasher));
    AvrXRunTask(TCB(switcher));
    AvrXRunTask(TCB(Monitor));

    AvrXSetSemaphore(&EEPromMutex); // Enables EEPROM access in monitor
    InitSerialIO(UBRR_INIT);    // Initialize USART baud rate generator

    Epilog();                   // Switch from AvrX Stack to first task
    return 0;
}

