/*
	Example code for setting up and using timers

	Two tasks each waiting on a timer, flashing LED's on 
	the evaluation board.  See "hardware.h" for LED definitions

	Pay attention to the AvrXStack defined in the makefile.
	this typically matches the stack passed to main().
*/
#include <io.h>             // GCC I/O definitions and other good stuff
#include <signal.h>
#include "AvrX.h"
#include "hardware.h"

void InitSerialIO(uint8_t);     // From Avrx.a (debug monitor)

AVRX_TASK(Monitor, 20, 0);      // External Task: Debug Monitor

TimerControlBlock   timer1,     // Declare the control blocks needed for timers
                    timer2;

#define TCNT0_INIT (0xFF-CPUCLK/256/TICKRATE)
#define TMC8_CK256 (1<<CS02)
#define UBRR_INIT (CPUCLK/(16*BAUDRATE)-1)
/*
 Timer 0 Overflow Interrupt Handler

 Prototypical Interrupt handler:
 . Switch to kernel context
 . handle interrupt
 . switch back to interrupted context.
 */
AVRX_SIGINT(SIG_OVERFLOW0)
{
    IntProlog();                // Switch to kernel stack/context
    outp(TCNT0_INIT, TCNT0);    // Reset timer overflow count
    AvrXTimerHandler();         // Call Time queue manager
    Epilog();                   // Return to tasks
}
/*
 Task 1 simply flashes the light off for 1/5 second and then on for 4/5th
 for a 1 second cycle time.
 */
AVRX_TASKDEF(task1, 6, 3)
{
    while (1)
    {
        AvrXStartTimer(&timer1, 800);       // 800 ms delay
        AvrXWaitTimer(&timer1);
        outp(inp(PORTB) ^ 0x01, PORTB);     // Toggle LED
        AvrXStartTimer(&timer1, 200);       // 200 ms delay
        AvrXWaitTimer(&timer1);
        outp(inp(PORTB) ^ 0x01, PORTB);     // Toggle LED
    }
}
/*
 Task 2 cycles the light on/off over 4 seconds.  It uses a simplified
 form of the delay API
 */
AVRX_TASKDEF(task2, 6, 2)
{
    while (1)
    {
        AvrXDelay(&timer2, 2000);           // 2 second delay
        outp(inp(PORTB) ^ 0x02, PORTB);     // Toggle LED
    }
}

void main(void)                 // Main runs under the AvrX Stack
{
    outp((1<<SE) , MCUCR);      // Enable "Sleep" instruction

    outp(TCNT0_INIT, TCNT0);
    outp(TMC8_CK256 , TCCR0);   // Set up Timer0 for CLK/256 rate
    outp((1<<TOIE0), TIMSK);    // Enable0 Timer overflow interrupt

    outp(-1, DDRB);		        // Make PORTB output and 
    outp(-1, PORTB);            // drive high (LEDs off)

    AvrXRunTask(&task1Tcb);	    // Init and run our tasks
    AvrXRunTask(&task2Tcb);
    AvrXRunTask(&MonitorTcb);

    InitSerialIO(UBRR_INIT);    // Initialize USART baud rate generator

    Epilog();                   // Switch from AvrX Stack to first task
}

