#include "hardware.inc"
#include <io.h>
#include "avrx.h"
/*
 Basic Tasking Tests #2

 Tests each semaphore call (not aliases)

 With Basic Test 1, this test pretty much covers basic tasking and semaphore
 operation.

 Runs through Set/Reset of Semaphores with things waiting and
 not waiting.  Tests actual state of semaphores to verify their action

 Test out other tasking stuff:
    AvrXTerminate
    AvrXResume      // Shouldn't resume a terminated task
    AvrXSuspend
    AvrXResetSemaphore

 NOTE: This test will generate a bunch of casting errors (warnings) when
 compiled.  No big deal...
 */
AVRX_TASK(task1, 20, 3);
AVRX_TASK(task2, 20, 2);    // Higher priority so they run 
AVRX_TASK(task3, 20, 2);    // to completion before #1 gets control again
AVRX_MUTEX(Simple);
AVRX_TIMER(Object);
AVRX_MUTEX(Done);


void main(void)
{
    AvrXRunTask(&task1Tcb);

    Epilog();           // Switch from AvrX Stack to first task
}

void task1(void) __attribute__ ((naked));
void task1(void)
{
    outp(-1, DDRB);     // Initialize LED port
    outp(-1, PORTB);

    while(1)
    {
        AvrXRunTask(&task2Tcb);       // Task blocks on Simple semaphore
        if (AvrXTestSemaphore(&Simple) != (Mutex)2)
            AvrXHalt();
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        
        AvrXSuspend(&task2Pid);     // Still Queued, but marked for suspension
        AvrXSetSemaphore(&Simple);  // Ok, released, but now suspended
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        if (AvrXTestSemaphore(&Simple) != (Mutex)_PEND)
            AvrXHalt();
        
        AvrXResume(&task2Pid);      // Ok, done and waiting on Simple again.
        if (AvrXTestSemaphore(&Done) != (Mutex)_DONE)
            AvrXHalt();
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND) // Test resets _DONE to _PEND
            AvrXHalt();
        if (AvrXTestSemaphore(&Simple) != (Mutex)2)
            AvrXHalt();
        
        AvrXSuspend(&task2Pid);     // Still Queued, but marked for suspension
        AvrXSetSemaphore(&Simple);  // Now _PEND
        AvrXSetSemaphore(&Simple);  // Now _DONE
        AvrXResetSemaphore(&Simple);// Now _PEND
        if (AvrXTestSemaphore(&Simple) != _PEND)
            AvrXHalt();
        AvrXResume(&task2Pid);      // Now blocked on Simple
        AvrXResetSemaphore(&Done);   // Should have been _DONE
        AvrXResetSemaphore(&Simple); // this should do nothing
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        if (AvrXTestSemaphore(&Simple) != (Mutex)2)
            AvrXHalt();
        AvrXTerminate(&task2Pid);   // Mark task for termination
        AvrXSetSemaphore(&Simple);  // Release from Semaphore so it terminates.
        AvrXResume(&task2Pid);      // This should have no effect
    
        //
        // Now do it all again but with an Object...
        // use "Timer" since there is no data type "object"
        //
            
        AvrXRunTask(&task3Tcb);       // Task blocks on Simple semaphore
        if (AvrXTestTimer(&Object) != (Mutex)2)
            AvrXHalt();
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        
        AvrXSuspend(&task3Pid);     // Still Queued, but marked for suspension
        AvrXSetSemaphore(&Object.semaphore);  // Ok, released, but now suspended
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        if (AvrXTestTimer(&Object) != (Mutex)_PEND)
            AvrXHalt();
        
        AvrXResume(&task3Pid);      // Ok, done and waiting on Object again.
        if (AvrXTestSemaphore(&Done) != (Mutex)_DONE)
            AvrXHalt();
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND) // Test resets _DONE to _PEND
            AvrXHalt();
        if (AvrXTestTimer(&Object) != (Mutex)2)
            AvrXHalt();
        
        AvrXSuspend(&task3Pid);     // Still Queued, but marked for suspension
        AvrXSetSemaphore(&Object.semaphore);  // Now _PEND
        AvrXSetSemaphore(&Object.semaphore);  // Now _DONE
        AvrXResetSemaphore(&Object.semaphore);// Now _PEND
        if (AvrXTestTimer(&Object) != _PEND)
            AvrXHalt();
        AvrXResume(&task3Pid);      // Now blocked on Object
        AvrXResetSemaphore(&Done);   // Should have been _DONE
        AvrXResetSemaphore(&Object.semaphore); // this should do nothing
        if (AvrXTestSemaphore(&Done) != (Mutex)_PEND)
            AvrXHalt();
        if (AvrXTestTimer(&Object) != (Mutex)2)    // Should be waiting on semaphore
            AvrXHalt();
        AvrXTerminate(&task3Pid);
        AvrXSetSemaphore(&Object.semaphore);
        AvrXResume(&task3Pid);
            
        outp(inp(PORTB) ^ 1, PORTB);          // Toggle bits
    }
}

void task2(void) __attribute__ ((naked));
void task2(void)
{
    while(1)
    {
        AvrXWaitSemaphore(&Simple);
        outp(inp(PORTB) ^ 2, PORTB);          // Toggle bits
        AvrXSetSemaphore(&Done);
    }
}

void task3(void) __attribute__ ((naked));
void task3(void)
{
    while(1)
    {
        AvrXWaitTimer(&Object);
        outp(inp(PORTB) ^ 4, PORTB);          // Toggle bits
        AvrXSetSemaphore(&Done);
    }
}

