#ifndef __AVRXINC
#define __AVRXINC
/*

Copyright 1998-2000 Larry Barello

Author: Larry Barello
        larry@barello.net
*/

#include "io-avr.h"

#ifndef BV
# define BV(A) (1<<(A))
#endif

/*
 Frame Pointer Offsets for various registers after the return from
 a call to _Prolog
*/
#define _RetLo  33      /* Return address on stack big endian */
#define _R31    33      /* Registers saved in order (little endian) */
#define _RetHi  32
#define _R30    32
#define _R29    31
#define _R28    30
#define _R27    29
#define _R26    28
#define _R25    27
#define _R24    26
#define _R23    25
#define _R22    24
#define _R0     2
#define _SREG   1       /* SP points below last saved item */


/* ******** TCB (Task Control Block) offsets */

#define TaskSP          0       /* Stack pointer */
#define TaskMain        2       /* Entry point of process */
#define TaskPid         4       /* Pointer to Process ID */
#define TaskFlags       6       /* Priority */

#define TaskSz  7

/* All objects have same link list structure */

#define NextL 1
#define NextH 0

/******** PID (Process ID) block offsets */

#define PidNext         0       /* Next item on list (semaphore, run) */
#define PidState        2       /* Upper Nibble: Task flags, Lower Nibble :Priority */
#define PidPriority     3
#define PidSP           4       /* Context Pointer */
#ifdef SINGLESTEPSUPPORT
#   define bp1          6       /* Make room for two breakpoints/task */
#   define bp2          8
#   define PidSz        10
#else
#   define PidSz        6
#endif

/* ******* PID (Process ID) BLOCK BIT DEFINITIONS ******* */

/* PID.State  */

#define IdleBit         4       /* Dead Task, don't schedule, resume or step */
#define SuspendBit      5       /* Mark task for suspension (may be blocked elsewhere) */
#define SuspendedBit    6       /* Mark task suspended (it was removed from the run queue) */
#define SingleStep      7       /* Enable single step debug support */

/*+ --------------------------------------------------
SEMAPHORE BIT DEFINITIONS

 General Semaphore Flags.  When a task is waiting on the
 semaphore, the semaphore is set to the TASK ID.  Otherwise
 _PEND means the semaphore is set and _DONE means it has
 been triggered.  A task will only block if it tries to 
 grab a _PEND semaphore.  Semaphores are initialized in the
 _PEND state and are left in the _PEND state after each 
 complete operation.  Processes are allowed to queue on
 a semaphore.  This effectively make the semaphore a Mutex
 when it has been "set" prior to the first call.  When a process
 done with the mutex, it releases it by "setting" it again.
-*/
#define _PEND           0
#define _DONE           1

#define MtxSz           2

#define MUTEX(A) \
        .global A\
        .comm   A, 2

/*+ --------------------------------------------------

 Events

 Events replace PID's in semaphores.  The kernel recognizes
 semaphore values (contents) < 0x60 as events, indexes into
 a table and calls a routine.  Macros and segments are used
 to build the table.

 */
#define TIMERMESSAGE_EV 2
#define PERIODICTIMER_EV 3      /* Not implemented */
#define _LASTEV 0x5F

/* Timers */

#define TcbNext         0       /* Pointer in linked list */
#define TcbSemaphore    2       /* Associated semaphore */
#define TcbCount        4       /* Timer ticks till expired */
#define TcbQueue        6
#define TcbSz           6       /* Primitive Timer */
#define TmbSz           8       /* Timer Message */

#define TIMER(A) \
        .global A \
        .comm   A, TcbSz
        
#define TIMERMESSAGE(A) \
        .global A \
        .comm   A, TmbSz
        
/* Message Queue */

#define MsqSz           4       /* Head of message queue */

#define MsqMessage      0       /* Head of list of messages */
#define MsqPid          2       /* Head of list of waiting processes */

#define QcbSz           4       /* Queue Block Size (No data) */

#define QcbNext         0
#define QcbSemaphore    2       /* Return Receipt Semaphore */
#define QcbData         4       /* pointer to data/or data */

#define MESSAGE(A, B) \
        seg eram.avrxdata.messages.A \
public A ds.b    QcbSz + (B)

#define MESSAGEQ(A) \
        .global A\
        .comm   A, MsqSz

/* Fifos */

#define FifoSz 4        /* Size of AvrX Fifo Data Struct */

#define WriteSemaphore 0
#define ReadSemaphore 2
#define fBuffer 4

/* User Section */

#define f_buffersz 4    /* Size of buffer header */

#define f_Size 0
#define f_Count 1
#define f_In 2
#define f_Out 3
#define f_Data 4        /* Variable length */

#define FIFO(A, B, Size) \
        .comm   A, FifoSz \
        .comm   B, (Size+f_Buffersz)

#ifndef AVRXCHEADER
/*+ --------------------------------------------------
 Handy Macros

 DEBUG statements are diagnostics for twiddling and output port bit
 to reflect the I bit state (interrupt latency) and when AvrX is in
 the Idle loop
-*/
#ifdef  DEBUG           /* DEBUG is set to a PORT{ABCD} value (for twiddling output bits) */

#ifndef ILED
#define ILED            PB0
#endif

#ifndef IDLELED
#define IDLELED         PB4
#endif

#define DEBUG_DDR       (DEBUG-1)

#endif  /* DEBUG */

        .macro  Breakpoint
#ifdef  DEBUG
        rcall   AvrXBreakpoint
#endif  /* DEBUG */
        .endm

        .macro  BeginInterrupt
#ifdef  DEBUG
        cbi     DEBUG, ILED
        sbi     DEBUG, IDLELED  /* Obviously not idle... */
#endif  /* DEBUG */
        .endm

        .macro  EndInterrupt
#ifdef  DEBUG
        sbi     DEBUG, ILED
#endif  /* DEBUG */
        reti
        .endm
        
        .macro BeginCritical
#ifdef  DEBUG
        cbi     DEBUG, ILED
#endif  /* DEBUG */
        cli     ;
        .endm
        
        .macro  EndCritical
#ifdef  DEBUG
        sbi     DEBUG, ILED
#endif  /* DEBUG */
        sei
        .endm
        
#define EndCriticalReturn EndInterrupt

        .macro  BeginIdle
snore:
#ifdef  DEBUG
        cbi     DEBUG, IDLELED
#endif  /* DEBUG */
        .endm
        
        .macro  EndIdle
snooze:
#ifdef  DEBUG 
        sbi     DEBUG, IDLELED
#endif  /* DEBUG */
        .endm
        
        .macro  INIT_DEBUG
#ifdef  DEBUG
        ldi     Xl, (1<<ILED | 1<<IDLELED)
        out     DEBUG_DDR, Xl
        out     DEBUG, Xl 
#endif  /* DEBUG */
        .endm

/*+ --------------------------------------------------
 Handy macros for common assembly chores
-*/
        .macro  SignExtend Dest, Source
        clr     Dest
        sbrc    Source, 7
        com     Dest
        .endm

        .macro  addi reg, val
        subi    reg, -(val)
        .endm

        .macro  adci reg, val
        sbci    reg, -(val)
        .endm

        .macro  tstw reg
        adiw    reg, 0
        .endm
        
        .macro  negw regh, regl
        com     regh
        neg     regl
        sbci    regh, -1
        .endm
/*
 Use this macro rather than a call to _Prolog, see
 version notes in AvrX.asm
*/
        .macro  AVRX_Prolog
        BeginCritical
        rcall   IntProlog
        EndCritical
        .endm
/*
 These register definitions are just handy aliases for the
 various index and word math registers
 */
#define		Wl 		R24
#define		Wh 		R25
#define		Xl 		R26
#define		Xh 		R27
#define		Yl 		R28
#define		Yh 		R29
#define		Zl 		R30
#define		Zh 		R31
 
#endif /*AVRXCHEADER */
#endif  /* __AVRXINC */
