#ifndef AVRXCHEADER
#define AVRXCHEADER
#include "avrx.inc"
#include <inttypes.h>
/*
avrx.h

Function prototypes for AvrX C library

Copyright 1998-2000 Larry Barello

Author: Larry Barello
        larry@barello.net

Revision History
    Aug 11 - lb
        The data structure for ProcessID used the macro define
        DEBUG to conditionally enable the breakpoint structure.  That should
        have been SINGLESTEPSUPPORT to match the assembly code.

        The data structure MessageControlBlock member Data was removed.  This
        is a vestigal data location.  See example "MessageAndData.c" for an
        example of how to extend the MCB with user data structures.

*/
#define FLASH  __attribute__ ((progmem))
#define EEPROM __attribute__ ((section(".eeprom")))
#define NAKED  __attribute__ ((naked))

/*
    The process id is a chunk of eram that contains the state
    of a process.
*/
typedef struct ProcessID
{
    struct ProcessID  *next;
    uint8_t flags, priority;
    void *ContextPointer;
#ifdef SINGLESTEPSUPPORT
    unsigned char *bp1;
    unsigned char *bp2;
#endif
}
* pProcessID, ProcessID;

#define NOMESSAGE ((pMessageControlBlock)0)
#define NOTIMER ((pTimerControlBlock)0)
#define NOPID ((pProcessID)0)
#define SEM_PEND ((Mutex)_PEND)
#define SEM_DONE ((Mutex)_DONE)
#define SEM_WAIT ((Mutex)2)     // AvrXTestSemaphore(): Something waiting

/*
 Mutex semaphores are a simple linked list of waiting
 processes.  The mutex may have the following values:

 SEM_PEND         // Semaphore is reset waiting for a signal
 SEM_DONE         // Semaphore has been triggered.
                  // Any other value is the address of a processID
*/
typedef pProcessID Mutex, *pMutex;     /* A mutex is basically a pointer to a process */

#define AVRX_MUTEX(A)\
        Mutex A
        
extern void AvrXSetSemaphore(pMutex);
extern void AvrXIntSetSemaphore(pMutex);
extern void AvrXWaitSemaphore(pMutex);
extern Mutex AvrXTestSemaphore(pMutex);
extern Mutex AvrXIntTestSemaphore(pMutex);
extern void AvrXResetSemaphore(pMutex);
extern void AvrXResetObjectSemaphore(pMutex);
/*
    Message Queues are really an extension of Semaphore
    queue.  They can be used for mutual exlusion (have a
    message be the baton) or as a general purpose message
    passing system.  This is needed for tasks that must block
    waiting for several things at once (timer, interrupt, etc)
*/
typedef struct MessageControlBlock
{
    struct MessageControlBlock *next;
    Mutex semaphore;
/*    uint8_t data;     08/11/00 lb */
}
* pMessageControlBlock, MessageControlBlock;

// Lame macro...  Should just declare directly.

#define AVRX_MESSAGE(A) /* 08/11/00 lb */\
        MessageControlBlock A

typedef struct MessageQueue
{
    pMessageControlBlock message;    /* List of messages */
    pProcessID pid;        /* List of processes */
}
* pMessageQueue, MessageQueue;

#define AVRX_MESSAGEQ(A)\
        MessageQueue A

extern pMessageControlBlock AvrXRecvMessage(pMessageQueue);
extern pMessageControlBlock AvrXWaitMessage(pMessageQueue);
extern void AvrXSendMessage(pMessageQueue, pMessageControlBlock);
extern void AvrXIntSendMessage(pMessageQueue, pMessageControlBlock);
extern void AvrXAckMessage(pMessageControlBlock);
extern void AvrXWaitMessageAck(pMessageControlBlock);
extern Mutex AvrXTestMessageAck(pMessageControlBlock);
/*
    The timer queue manager is a service run in kernel mode and is tuned
    to minimize interrupt latency while queueing, tracking and dequeuing
    timers
*/
typedef struct TimerControlBlock
{
    struct TimerControlBlock *next;
    Mutex semaphore;
    unsigned short count;
}
* pTimerControlBlock, TimerControlBlock;
/*
    A special version of timers that send messages rather than firing
    a semaphore.
*/
typedef struct TimerMessageBlock
{
    union
    {
        struct MessageControlBlock mcb;
        struct TimerControlBlock tcb;
    } u;
    struct MessageQueue *queue;
}
* pTimerMessageBlock, TimerMessageBlock;
    

#define AVRX_TIMER(A) TimerControlBlock A

extern void AvrXStartTimer(pTimerControlBlock, unsigned);
extern pTimerControlBlock AvrXCancelTimer(pTimerControlBlock);
extern void AvrXDelay(pTimerControlBlock, unsigned);
extern void AvrXWaitTimer(pTimerControlBlock);
extern Mutex AvrXTestTimer(pTimerControlBlock);

extern void AvrXTimerHandler(void);    // Kernel Funtion called by timer ISR

// Special versions of timer queue elements that get sent
// to a message queue when expired.

extern void AvrXStartTimerMessage(pTimerMessageBlock timer, uint16_t timeout, pMessageQueue queue);
extern pMessageControlBlock AvrXCancelTimerMessage(pTimerMessageBlock, pMessageQueue);

/*
   The Task Control Block contains all the information needed
   to initialize and run a task.  It is stored in FLASH and is
   used only by AvrXInitTask()
*/
typedef struct
{
        void *stack;            // Start of stack (top address-1)
        void (*start) (void);   // Entry point of code
        pProcessID pid;         // Pointer to Process ID block
        uint8_t priority;       // Priority of task (0-255)
}
TaskControlBlock FLASH;
/*
    A series of macro's to ease the declaration of tasks
    and access to the resulting data structures.

AVRX_TASK(start, stacksz, priority)
	Declare task data structures and forward reference to task
AVRX_TASKDEF(start, stacksz, priority)
	Declare task data structure and the top level C
	declaration (AVRX_TASK + C function declaration)
AVRX_SIGINT(vector)
	Declare the top level C declaration for an 
	interrupt handler
AVRX_EXTERNTASK(start)
	Declare external task data structures 
PID(start)
	Return the pointer to the task PID
TCB(start)
	Return the pointer to the task TCB
*/

#define MINCONTEXT 35           // 32 registers, return address and SREG

#define AVRX_TASK(start, stacksz, priority)	\
  void start(void) NAKED; \
  ProcessID start ## Pid; \
  char start ## Stk [stacksz + MINCONTEXT] ; \
  TaskControlBlock start ## Tcb = \
  { \
        &start##Stk[sizeof(start##Stk)-1] , \
        start, \
        &start##Pid, \
        priority \
  }
  
#define AVRX_TASKDEF(start, stacksz, priority) \
  AVRX_TASK(start, stacksz, priority); \
  void start(void)
  
#define AVRX_SIGINT(vector)\
  void vector(void) NAKED;\
  void vector(void)

#define PID(start) &start##Pid
#define TCB(start) (&start##Tcb)

#define AVRX_EXTERNTASK(start) \
  void start(void); \
  extern TaskControlBlock start##Tcb; \
  extern ProcessID start##Pid

extern void AvrXRunTask(const TaskControlBlock *);
extern unsigned char AvrXInitTask(const TaskControlBlock *);

extern void AvrXResume(pProcessID);
extern void AvrXSuspend(pProcessID);
extern void AvrXBreakPoint(pProcessID);
extern uint8_t AvrXSingleStep(pProcessID);
extern uint8_t AvrXSingleStepNext(pProcessID);

extern void AvrXTerminate(pProcessID);
extern void AvrXTaskExit(void);
extern void AvrXHalt(void);     // Halt Processor (error only)

extern void AvrXWaitTask(pProcessID);
extern Mutex AvrXTestPid(pProcessID);

uint8_t AvrXPriority(pProcessID);
uint8_t AvrXChangePriority(pProcessID, uint8_t);
pProcessID AvrXSelf(void);
extern void IntProlog(void);
/*
  Can't "call" Epilog from kernel mode as it messes up the stack!
  Ok to do transitioning from user to kernel or back since the stack
  gets reset...
 */
// void Epilog(void) __attribute__ ((__noreturn__));
extern inline void Epilog(void) { asm("\trjmp     _Epilog\n"); }

extern inline void BeginCritical(void) { asm volatile("cli\n"); }
extern inline void EndCritical(void) { asm volatile("sei\n"); }

/*
 AvrX provides some EEPROM access routines that
 control access to the hardware via a semaphore
 This semaphore needs to be "set" prior to using
 the access routines.
 */
extern Mutex EEPromMutex;

#endif /* AVRXCHEADER */
