#include        "AvrX.inc"
/*
        Copyright 1999-2000, Larry Barello
        larry@barello.net

*/

        .section .text.avrx.moretasking

/*+
; --------------------------------------------------
; void AvrXWaitObjectSemaphore(pMutex)
; void AvrXWaitTimer:
; void AvrXWaitMessageAck:
; void AvrXWaitTask:
;
; Standard system objects have a semaphore as their
; second byte.  This call waits on that.
;
; PASSED:       R25:R24 = Object
; RETURNS:      
; USES:         Z, X;       
; CALLS:
; ASSUMES:
; NOTES:        - 6/28/00 Changed code to only do IntProlog/Epilog
;               if Semaphore is _PEND or > _DONE
;               - Must stay critical through entire routine so we
;               don't "append" ourselves to a "DONE" semaphore
;               - Because of above, can assume the running code is
;               at the top of the run queue (change code later..)
;
;               Trivial case ~20 cycles
-*/
        .global AvrXWaitObjectSemaphore
        .global AvrXWaitTimer
        .global AvrXWaitMessageAck
        .global AvrXWaitTask
        .func   AvrXWaitObjectSemaphore
        
AvrXWaitTimer:
AvrXWaitMessageAck:
AvrXWaitTask:
AvrXWaitObjectSemaphore:
        adiw    R24, 2          ; Fall through to WaitSemaphore
        .endfunc
/*+
; -----------------------------------------------
; void AvrXWaitSemaphore(pMutex)
;
; Tasks will queue up for a Semaphore, thus implementing 
; a Mutual Exclusion Semaphore.  For just plain old signaling
; only one task should wait on a semaphore.
;
; PASSED:       R25:R24 =  Semaphore
; USES:         Z, X
; Returns:      
; STACK:        One Context
;
; Notes:        Only called from user mode, may block.
-*/

        .global AvrXWaitSemaphore
        .func   AvrXWaitSemaphore
AvrXWaitSemaphore:
        mov     Zl, R24
        mov     Zh, R25

        BeginCritical           ; Must stay critical through entire routine

        ldd     Xl, Z+NextL     ; Get semaphore
        ldd     Xh, Z+NextH
        sbiw    Xl, _DONE       ; Test trivial case of _DONE
        brne    aws01

        std     Z+NextL, Xl     ; Reset Semaphore to _PEND
        std     Z+NextH, Xh
        EndCriticalReturn       ; and return
aws01:
        adiw    Xl, _DONE       ; Restore semaphore

        rcall   IntProlog       ; Do task switch (ints disabled)

        ; With new code, we *can* assume we are at the top of the run queue
        
        lds     R23, Running
        lds     R22, Running+1
        ldi     Zl, lo8(RunQueue)
        ldi     Zh, hi8(RunQueue)
        rcall   _RemoveObject   ; Remove ourself from the run queue
        mov     Zl, R24
        mov     Zh, R25
        rcall   _AppendObject   ; Append ourselves to the Semaphore

        rjmp    _Epilog
        .endfunc
/*+
; -----------------------------------------------
; void AvrXSetSemaphore(pMutex)
; void AvrXSetObjectSemaphore(pMythicalObject)
; void AvrXAckMessage(pMessageControlBlock)
;
; Sets (signals) an regular or object semaphore
;
; Signals a semaphore.  If semaphore is _PEND or _DONE, just mark
; DONE and return.  If a task is waiting, run it and set the semaphore 
; to the _PEND state.  
;
; Second Entry point is for Interrupt routines that want to Signal
; a semaphore.  Assumes context already saved.
;
; PASSED:       R25:R24 = Semaphore
; RETURNS:
; USES:         X, Z, R25:R24
; CALLS:        AvrXIntSetSemaphore
; STACK:
; SIDE EFFECTS: 
-*/
        .global AvrXAckMessage
        .global AvrXSetObjectSemaphore
        .func   AvrXSetObjectSemaphore
AvrXAckMessage:
AvrXSetObjectSemaphore:
        adiw    R24, 2          ; Fall through to SetSemaphore
        .endfunc
        
        .global AvrXSetSemaphore
        .func AvrXSetSemaphore
AvrXSetSemaphore:
        rcall   AvrXIntSetSemaphore
        tst     R24
        breq    1f
        ret
1:
        BeginCritical           ; Do task switch with interrupts off
        rcall   IntProlog
        EndCritical             ; Allow (any) pending interrupt to be serviced.
        rjmp    _Epilog
        
        .endfunc                
/*+
; -----------------------------------------------
; (void) AvrXIntSetSemaphore(pMutex)
; (void) AvrXIntSetObjectSemaphore(pMutex)
;
; Entry point for interrupt handlers (or code already switched into
; the system stack)
;
; PASSED:       R25:R24 = Semaphore
; RETURNS:      R24 != 0, nothing queued
;                   == 0, task queued
; USES:         Z, X, R25:R24
; STACK:        2
; SIDE EFFECTS:
; NOTES:        Doesn't accumulate sets.  If semaphore already set, the
;               just ignore it (not a counting semaphore)
-*/
        .global AvrXIntSetObjectSemaphore
        .func   AvrXIntSetObjectSemaphore
AvrXIntSetObjectSemaphore:      ; This one here because so small
        adiw    R24, 2
        .endfunc

        .global AvrXIntSetSemaphore
        .func   AvrXIntSetSemaphore
AvrXIntSetSemaphore:            ; Entry point for interrupt exit routines.
        mov     Zl, R24
        mov     Zh, R25
        ldi     R24, lo8(_DONE)
        ldi     R25, hi8(_DONE)

        BeginCritical

        ldd     Xh, Z+NextH
        ldd     Xl, Z+NextL
        cp      R24, Xl 
        cpc     R25, Xh
        brlo    aiss00           ; Trivial case of _PEND or _DONE
        std     Z+NextL, R24
        std     Z+NextH, R25     ; Set to _DONE

        EndCriticalReturn
        
aiss00:
        ldi     R24, lo8(_LASTEV); Check for valid semaphore (e.g. >= $60)
;       ldi     R25, hi8(_LASTEV)
        cp      R24, Xl 
        cpc     R25, Xh
        brsh    BogusSemaphore

        rcall   _RemoveObjectAt ; Z->Prev, X->Next (Object)
        
        EndCritical
        
        mov     R25, Xh
        mov     R24, Xl
        rcall   _QueuePid       ; X = Pid
        clr     R24
//        ret

BogusSemaphore:
        EndCriticalReturn
        .endfunc
/*+
; -------------------------------------------------
; _AppendObject
;
; Appends an object to a queue
;
; PASSED:       Z = Queue head
;               R23:R22 = Object
; RETURNS:
; USES:         Z, X & Flags
; CALLS:
; ASSUMES:      Null terminated list; 
; NOTES:        
-*/
        .global _AppendObject
_ao00:
        mov     Zl, Xl
        mov     Zh, Xh
        .func   _AppendObject
_AppendObject:
        ldd     Xl, Z+NextL
        ldd     Xh, Z+NextH
        adiw    Xl, 0
        brne    _ao00           ; Zip to end of list
        
        std     Z+NextH, R23
        std     Z+NextL, R22    ; Prev->Next = Object
        mov     Zh, R23
        mov     Zl, R22
        std     Z+NextH, Xh     ; Object->Next = Next
        std     Z+NextL, Xl
        ret
        .endfunc
/*+
; --------------------------------------------------
; _RemoveObject
;
; Removes an object from a queue.  Returns 0 if failed.
;
; PASSED:       Z = Queue Head
;               R23:R22 = Object
; RETURNS:      X       = Object removed, 0 if failed
;               Z       = Next object in list
; USES:         Z, X, Flags
; CALLS:
; ASSUMES:      Called within a critical section
; NOTES:        Z are side effects used else where in AvrX
-*/
        .global _RemoveObject
_ro00:
        mov     Zh, Xh
        mov     Zl, Xl
        .func _RemoveObject
_RemoveObject:
        ldd     Xl, Z+NextL
        ldd     Xh, Z+NextH
        cp      R22, Xl
        cpc     R23, Xh
        breq    _RemoveObjectAt ; Match, we got it.
        adiw    Xl, 0
        brne    _ro00           ; Walk the list
        ret                     ; End of list
        .endfunc
        
/*+
; --------------------------------------------------
; pQueue _RemoveFirstObject(pQueue)
;
; Removes the first item in a queue. Return Zero if empty.
;
; PASSED:       Z       = Queue head
; RETURNS:      X       = Object = 0 if list empty
;                       If X != 0, then R23:22 = X
;               Z       = Next Item
; USES:         R23:R22 & Flags
; CALLS:
; ASSUMES:      Null terminated list
; NOTES:        Interrupts NOT disabled.  Must disable if possibility
;               of something being insterted at head of queue
-*/
        .global _RemoveFirstObject
        .func _RemoveFirstObject
_RemoveFirstObject:
        ldd     Xl, Z+NextL
        ldd     Xh, Z+NextH     ; X = pointer to object
        adiw    Xl, 0
        brne    _RemoveObjectAt
        ret
        .endfunc
        
/*+
; --------------------------------------------------
; pQueue _RemoveObjectAt(pQueue)
;
; Removes the  item in a queue. 
;
; PASSED:       Z       = Prev
; RETURNS:      X       = Object (R23:R22)
;               Z       = Next Item
; USES:         R23:R22 & Flags
; CALLS:
; ASSUMES:      Null terminated list
; NOTES:        Interrupts NOT disabled.  Must disable if possibility
;               of something being insterted at head of queue
-*/
        .global _RemoveObjectAt
        .func _RemoveObjectAt

_RemoveObjectAt:                ; Z = Prev, X = Object
        ld      R23, X+
        ld      R22, X
        std     Z+NextH, R23
        std     Z+NextL, R22    ; Prev->Next = Object->Next
        mov     Zl, R22         ; Return Next in Z
        mov     Zh, R23
        clr     R22
        st      X, R22    ; Object->Next = 0
        st      -X, R22
        mov     R23, Xh
        mov     R22, Xl
        ret
        .endfunc

