#include        "AvrX.inc"
/*
        Copyright 1999-2000, Larry Barello
        larry@barello.net

*/
;+
;----------------------------------------------------
; Critical Data Structures
;
; - RunQueue is the head of the run queue.
; - Running is current running task - once inside a Prolog/Epilog
;       section, the RunQueue might get re-arranged so we need to track
;       who we are.
; - Syslevel is a counter tracking how many times we have reentered
; the kernel, or, how many interrupts are nested (same thing).
;-
        .comm   RunQueue, 2     ; Head of the run queue
        .comm   Running, 2      ; Current running task
        .comm   SysLevel, 1

        .section  .text.avrx.tasking
;+
; --------------------------------------------------
; void IntProlog(void)
;
; Pushes entire register context onto the stack, returning a frame pointer
; to the saved context.  If running in user mode upon entry (SysLevel == 0xFF)
; then switches the stack to the kernel and stores the frame pointer in the
; current processes PID.
;
; PASSED:       Nothing
; RETURN:       Y = Frame Pointer (See AvrX.inc)
;               R0 = pid.state (Single Step Flag) or 0
; ASSUMES:      
; USES:         R14-R18, R26-R32, SysLevel
; CALLS:                
; NOTES:        Does not muck with R20-R25 so passed parameters from C to
;               AvrX functions can be accessed directly after a call to Prolog.
;-
	.global	IntProlog
	.func	IntProlog
IntProlog:                      ; 3 cycles
	push	R29
	push	R28
	push	R27
	push	R26
        push    R25
        push    R24
        push    R23
        push    R22
        push    R21
        push    R20
        push    R19
        push    R18
        push    R17
        push    R16
        push    R15
        push    R14
        push    R13
        push    R12
        push    R11
        push    R10
        push    R9
        push    R8
        push    R7
        push    R6
        push    R5
        push    R4
        push    R3
        push    R2
        push    R1
        push    R0
        in      R0, SREG
        push    R0

	in	Yl, SPL
	in	Yh, SPH         ; Grab frame pointer

	ldd	Xl, Y+_RetLo
	ldd	Xh, Y+_RetHi
        std     Y+_R30, Zl
        std     Y+_R31, Zh      ; Swap return address with Z

#ifdef SINGLESTEPSUPPORT
        clr     R0
#endif /* SINGLESTEPSUPPORT */
	
	lds	R16, SysLevel
	inc	R16
	sts	SysLevel, R16   ; (80)
	brne	AlreadyInKernel ; Check if first time in kernel

	lds	Zh, Running    ; Point to running task
	lds	Zl, Running+1
	std	Z+PidSP, Yh
	std	Z+PidSP+1, Yl

#ifdef  SINGLESTEPSUPPORT
        ldd     R17, Z+PidState ; Return to caller SingleStepFlag.
        sbrs    R17, SingleStep
        rjmp    SkipDequeue
        
        ldd     R19, Z+PidNext  ; Dequeue current task
        ldd     R18, Z+PidNext+1
        sts     RunQueue, R19
        sts     RunQueue+1, R18
        mov     R0, R17         ; Return to caller SingleStepFlag.
        cbr     R17, BV(SingleStep)
        std     Z+PidState, R17 ; Clear flag in PID
SkipDequeue:
#endif /* SINGLESTEPSUPPORT */

        ldi	Zl, lo8(AvrXStack)      
	ldi	Zh, hi8(AvrXStack)
	out	SPL, Zl
	out	SPH, Zh         ; Swap to kernel stack
AlreadyInKernel:                ; (85/102)
	clr     R1              ; R1 = __Zero__ for Avr-gcc
        mov     Zl, Xl
        mov     Zh, Xh
	ijmp			; Return to caller (89/106)
	.endfunc

/*+
; --------------------------------------------------
; _Epilog
;
; Restore previous context (kernel or user).
; If task has SingleStep flag set, then generate an interrupt
; before returning to the task.
;
; PASSED:       
; RETURN:       
; ASSUMES:      SysLevel >= 0 (running on kernel stack)
; USES:         SysLevel
; CALLS:                
-*/
	.func	_Epilog
	.global	_Epilog
_Epilog:
        BeginCritical
        lds     R16, SysLevel    ; Interrupts off..
        dec     R16
        brge    SkipTaskSwap

        lds     Zh, RunQueue
        lds     Zl, RunQueue+1
        adiw    Zl, 0
        breq    _IdleTask

        sts     Running, Zh
        sts     Running+1, Zl   ; Update current running task

        ldd     Yh, Z+PidSP
        ldd     Yl, Z+PidSP+1
        out     SPL, Yl
        out     SPH, Yh         ; 20 cycles
#ifdef  SINGLESTEPSUPPORT     
        ldd     R0, Z+PidState  ; X, Y and R0 available
        sbrc    R0, SingleStep
        rcall   GenerateInterrupt
#endif
SkipTaskSwap:                   ; 20/6
        sts     SysLevel, R16
        pop     R0
        out     SREG, R0
        pop     R0
        pop     R1
        pop     R2
        pop     R3
        pop     R4
        pop     R5
        pop     R6
        pop     R7
        pop     R8
        pop     R9
        pop     R10
        pop     R11
        pop     R12
        pop     R13
        pop     R14
        pop     R15
        pop     R16
        pop     R17
        pop     R18
        pop     R19
        pop     R20
        pop     R21
        pop     R22
        pop     R23
        pop     R24
        pop     R25
        pop     R26
        pop     R27
        pop     R28
        pop     R29
        pop     R30
        pop     R31
        EndCriticalReturn       ; 97/83 cycles with interrupts off

; Jump here if there are no entries in the _RunQueue.

_IdleTask:
        BeginIdle               ; Any interrupt will EndIdle
        EndCritical             ; Enable interrupts
        sleep                   ; Power Down..
        EndIdle
        rjmp    _Epilog
        .endfunc

#ifdef SINGLESTEPSUPPORT
/*+
; --------------------------------------------------
; Generate Interrupt (Single Step Support)
;
; PASSED:       Nothing
; USES:         X, Y and R0
; RETURNS:      Nothing
; NOTES:        Generates an interrupt.  Assumes global
;               interrupt flag is cleared
; 
;
-*/
        .func   GenerateInterrupt
        
GenerateInterrupt:
        ldi     Yl, 1<<CS00     ; 1x prescale
        ser     Xh
        in      R0, TCCR0       ; Previous prescaler
        in      Yh, TCNT0       ; Previous count
        out     TCNT0, Xh       ; Set maximum
        out     TCCR0, Yl       ; Interrupt generated NOW
        out     TCCR0, R0       ; Restore orignal prescaler
        out     TCNT0, Yh       ; Restore original count
        ret
        .endfunc
#endif /* SINGLESTEPSUPPORT */
        
/*+
; --------------------------------------------------
; AvrXInitProcess
; AvrXRunTask
;
; Initialize the PID based upon the TCB table.  Run Task
; "resumes" the process as well
; 
; PASSED:       R25:R24 = TCB
; RETURNS:      R25:R24 = PID
; USES:         X, Z and R0
; CALLS:
; NOTES:        TCB, Layout, below: 
;               2 PID
;               1 Flags
;               2 STACK
;               2 Entry
-*/
        .global AvrXRunTask
        .func   AvrXRunTask
AvrXRunTask:
        rcall   AvrXInitTask
        rjmp    AvrXResume
        .endfunc
        
        .global AvrXInitTask
        .func   AvrXInitTask
AvrXInitTask:
        mov     Zl, R24
        mov     Zh, R25
        
        rcall   lpm_inc         ; Get Stack pointer
        mov     Xl, R0
        rcall   lpm_inc         
        mov     Xh, R0
        
#ifdef  TASKEXIT
        ldi     R24, lo8_pm(AvrXTaskExit)
        st      X, R24          ; !Pre-dec is not a PUSH!
        ldi     R24, hi8_pm(AvrXTaskExit)
        st      -X, R24          ; Push task exit onto stack

        rcall   lpm_inc
        st      -X, R0
        rcall   lpm_inc
        st      -X, R0          ; Push task entry point
#else
        rcall   lpm_inc
        st      X, R0
        rcall   lpm_inc
        st      -X, R0          ; Push task entry point
#endif /* TASKEXIT */

        ldi     R22, 0
        ldi     R23, 33          ; R0-R31 and SREG get "pushed"
PushRegisters:
        st      -X, R22
        dec     R23             ; Fill rest of context with Zeros
        brne    PushRegisters   ; Decrement stack pointer
        
        sbiw    Xl, 1           ; because pre dec is not a push!

        rcall   lpm_inc
        mov     R24, R0          ; Get PID -> R23:22
        rcall   lpm_inc
        mov     R25, R0
        rcall   lpm_inc         ; Priority -> R0
        mov     Zl, R24
        mov     Zh, R25

        std     Z+PidSP+1, Xl
        std     Z+PidSP, Xh     ; Stash context frame pointer

        std     Z+PidPriority, R0
        ldi     Xl, BV(SuspendedBit) | BV(SuspendBit)
        std     Z+PidState, Xl  ; Mark state so we can "resume" pid
        std     Z+PidNext, R22
        std     Z+PidNext+1, R24; Clear pid.next
        ret

lpm_inc:                        ; Tiny helper...
        lpm
        adiw    Zl, 1
        ret
        .endfunc
/*+
; --------------------------------------------------
; AvrXResume
;
; Takes a PID and inserts it into the run queue.
;
; Uses two flags, Suspended and Suspend to determine if a task is
; actually suspended, as opposed to just being marked for suspension.
; Mark _SUSP to make QueuePid not queue, Mark _SUSPENDED when attempt
; to queue is made (e.g. not blocked on a semaphore).  This is because
; we can't "resume" something that is blocked on a semaphore, even though
; it is marked for suspension.
;
; PASSED:       R25:24 = PID to queue
; RETURNS:      R24 : 0 = Ok, 1 = Already queued (error)
; USES:         
; ASSUMES:
; NOTES:        
-*/
        .global AvrXResume
        .func AvrXResume
AvrXResume:                     ; User Entry Point
        AVRX_Prolog

        mov     Zl, R24
        mov     Zh, R25
        
        ldd     Xl, Z+PidState
        cbr     Xl, BV(SuspendBit)
        std     Z+PidState, Xl  ; clear Suspend flag

        sbrs    Xl, SuspendedBit
        rjmp    ar00
        
        cbr     Xl, BV(SuspendedBit)  ; If removed from run queue,
        std     Z+PidState, Xl  ; Clear flag and queue it up.
        
        rcall   _QueuePid       ; If found, then add to run queue
ar00:
        rjmp    _Epilog
        .endfunc
/*+
; --------------------------------------------------
; _QueuePid
;
; Takes a PID and inserts it into the run queue.  The run queue is sorted
; by priority.  Lower numbers go first.  If there are multiple tasks of equal 
; priority, then the new task is appended to the list of equals (round robin)
;
; PASSED:       R25:R24 = PID to queue
; RETURNS:      
; USES:         Z, X, R23:R22 and SREG, RunQueue
; ASSUMES:
; NOTES:        Returns with interrupts on.
;
-*/
        .global _QueuePid
        .func _QueuePid
_QueuePid:                              ; Kernel entry point only
        mov     Zl, R24
        mov     Zh, R25

        ldd     Xh, Z+PidState          ; Xh = Priority & Flags
        andi    Xh, (BV(SuspendBit) | BV(IdleBit)) ; if marked Suspended or idle
        brne    _qpSUSPEND
        
        ldd     R22, Z+PidPriority
        mov     Xl, Yl
        mov     Xh, Yh                  ; Preserve Y through this routine
        ldi     Yl, lo8(RunQueue)
        ldi     Yh, hi8(RunQueue)
        BeginCritical
_qp00:
        mov     Zl, Yl
        mov     Zh, Yh
        ldd     Yl, Z+PidNext+1
        ldd     Yh, Z+PidNext           ; Z = current, X = Next
        adiw    Yl, 0
        breq    _qp01                   ; End of queue, continue
        ldd     R23, Y+PidPriority
        cp      R22, R23
        brsh    _qp00                   ; Loop until pri > PID to queue
_qp01:
        std     Z+NextH, R25
        std     Z+NextL, R24    ; Prev->Next = Object
        mov     Zh, R25
        mov     Zl, R24
        std     Z+NextH, Yh     ; Object->Next = Next
        std     Z+NextL, Yl
        mov     Yl, Xl
        mov     Yh, Xh
        EndCriticalReturn

_qpSUSPEND:
        sbr     Xh, BV(SuspendedBit)  ; Mark suspended and return
        std     Z+PidState, Xh
        EndCriticalReturn
        .endfunc


