#include        "AvrX.inc"
#include        "serialio.inc"
/*
File:   Monitor.asm

Copyright 1998, 2000 Larry Barello

Author: Larry Barello
        larry@barello.net

        REVISION HISTORY

*/
;------------------------------------------------
; 
#ifdef USE_SRAM
        .lcomm  Address, 2
        .lcomm  EEAddress, 2
        .lcomm  ObjectId, 2     ; Default for Object ID
        .lcomm  DumpLength, 1   ; Default for dumplength
#else
        .equ  Address, 2        ; R2:3
        .equ  EEAddress, 4      ; R4:5
        .equ  ObjectId, 6       ; R6:7
        .equ  DumpLength, 8     ; R8
#endif
        .section .text.avrx.monitor

#define MonFlags R16            // Immediate register
#define WORDFLAG 0
#define EEFLAG 1
#define NL "\n"
#define CR "\r"

;-------------------------------------------------
; Monitor: program for examining and modifying
; registers, data, i/o and eeprom.
;
; NB    {opt, opt}      Must select one of the options
;       [opt, opt]      Optional input - can be ignored
;
; Command summary:
;
;       cmd                     Action
;       --------------------    ------
;       m{s,e}[w][aaaa]         Modify SRAM/EEram by byte/word starting at AAAA
;       d{s,e}[w][aaaa] [xx]    Dump XX Bytes/Words of Sram/EERam, starting at AAAA
;       h[XX]                   Halt process XX
;       s[XX]                   Step Process XX (run until next kernel entry)
;       n[XX]                   Step Process till next instruction
;       r[XX]                   Resume Process XX
;       x[XX]                   display context of XX (use 's' to modify)
;       i[xx]                   Signal (set) Semaphore
;       c[xx]                   cancel timer XX
;       p[xxxx]                 Initialize Process (AvrXInitTask)
;
;
; Stack Usage:
;       12 or more bytes 
;
; Register Usage:
;       R24-31  (mid size context save)
;
; Global data information
;
; 
Hello:  .ascii  CR
        .ascii  "AvrX Monitor V2", CR
        .ascii  CR
        .ascii  "Modify = m{s,e}[w][aaaa]", CR
        .ascii  "Dump   = d{s,e}[w][aaaa] [xx]", CR
        .ascii  "Halt   = h[oooo]", CR
        .ascii  "Step   = s[oooo]", CR
#ifdef  SINGLESTEPSUPPORT
        .ascii  "Next   = n[oooo]", CR
#endif
        .ascii  "Resume = r[oooo]", CR
        .ascii  "Context= x[oooo]", CR
        .ascii  "Signal = i[oooo]", CR
        .ascii  "Cancel = c[oooo]", CR
        .ascii  "Initask= p[aaaa]", CR
        .ascii  "Quit   = q", CR
        .ascii  "\000"
        .balign 2

        .global Monitor
        .func   Monitor
Monitor:
        ldi     R24, lo8(EEPromMutex)
        ldi     R25, hi8(EEPromMutex)
        rcall   AvrXSetSemaphore        ; Initialize EEProm Mutex
        
        ldi     Zl, lo8(Hello)
        ldi     Zh, hi8(Hello)
        rcall   _PrintString

        ldi     Zl, lo8(0x60)
        sts     Address+1, Zl
        ldi     Zh, hi8(0x60)
        sts     Address, Zh
        .endfunc

        .global DebugMonitor
        .func   DebugMonitor
DebugMonitor:
        rcall   PutCR
        ldi     TxByte, '>'
        rcall   PutChar 
        rcall   m_Parse
        rjmp    DebugMonitor

m_Parse:
        rcall   _GetChar
        cpi     RxByte, 'm'     ; Modify Memory
        breq    m_ModifyMemory
        cpi     RxByte, 'd'     ; Dump RAM
        breq    m_DumpMemory
        cpi     RxByte, 'h'     ; Halt (suspend) Process
        breq    m_HaltProcess
        cpi     RxByte, 's'     ; Step process
        breq    m_StepProcess
#ifdef  SINGLESTEPSUPPORT
        cpi     RxByte, 'n'     ; Next instruction
        breq    m_SingleStep
#endif
        cpi     RxByte, 'r'     ; Resume Process
        breq    m_ResumeProcess
        cpi     RxByte, 'i'     ; Signal Semaphore
        breq    m_SetSemaphore
        cpi     RxByte, 'c'     ; 
        breq    m_CancelTimer
        cpi     RxByte, 'x'
        breq    m_PrintProcessContext
        cpi     RxByte, 'p'
        breq    m_InitTask
        cpi     RxByte, 'q'
        breq    m_quit
m_err:
        ldi     TxByte, '?'
        rjmp    PutChar

m_quit:
        pop     Xl      ; Hack to allow user i/f to call monitor
        pop     Xl
        ret
        
m_ModifyMemory:
        rcall   GetFlagsAndAddress
        brtc    m_err
        rjmp    _ModifyMemory
m_DumpMemory:
        rcall   GetFlagsAndAddress
        brtc    m_err           ; Flags in Xh, Addr in Z
        cpi     RxByte, 0x0D    ; See if terminated with a CR
        breq    m1b
        rcall   GetDumpLength   ; Leave count in Xl
        brtc    m_err
m1b:
        lds     R19, DumpLength
        rjmp    _DumpMemory

m_InitTask:
        rcall   GetHexWord      ; This is a word address, not byte.
        brtc    m_err
        mov     R24, Yl
        mov     R25, Yh
        rjmp    AvrXInitTask

m_HaltProcess:
        rcall   GetObjectID
        brtc    m_err
        rjmp    AvrXSuspend

m_ResumeProcess:
        rcall   GetObjectID
        brtc    m_err
        rjmp    AvrXResume

m_CancelTimer:
        rcall   GetObjectID
        brtc    m_err
        rjmp    AvrXCancelTimer
        
m_PrintProcessContext:
        rcall   GetObjectID
        brtc    m_err
        rjmp    _PrintContext
        
m_SetSemaphore:
        rcall   GetObjectID
        brtc    m_err
        rjmp    AvrXSetSemaphore

m_StepProcess:
        rcall   GetObjectID
        brtc    m_err
        rcall   AvrXStepNext
        tst     R24
        brne    ms_NOTSUSPENDED
        ret

#ifdef  SINGLESTEPSUPPORT
        
m_SingleStep:
        rcall   GetObjectID
        brtc    m_err
        mov     Zl, R24
        mov     Zh, R25
NextStep:
        rcall   AvrXSingleStepNext      ; This will return after
        tst     R24
        brne    ms_NOTSUSPENDED         ; the step is done

        ldd     Yl, Z+PidSP+1           ; Y = Frame Pointer
        ldd     Yh, Z+PidSP

        ldd     Xl, Y+_RetLo+2          ; Y = Address of Next inst
        ldd     Yh, Y+_RetHi+2
        mov     Yl, Xl
        
        ldd     Xh, Z+bp1               ; Breakpoints are stored in 
        ldd     Xl, Z+bp1+1             ; the PID..
        adiw    Xl, 0
        breq    BreakPoint
        cp      Yl, Xl
        cpc     Yh, Xh
        breq    BreakPoint
        
        ldd     Xh, Z+bp2
        ldd     Xl, Z+bp2+1
        adiw    Xl, 0
        breq    NextStep                ; If we dont hit the breakpoint
        cp      Yl, Xl                  ; the monitor hangs here
        cpc     Yh, Xh
        brne    NextStep
        
BreakPoint:
        rjmp    PutHexWord              ; Print out our breakpoint address
#endif

ms_NOTSUSPENDED:
        rcall   PutCR
        rcall   PutHexByte
        ldi     Zl, lo8(NotSuspended)
        ldi     Zh, hi8(NotSuspended)
        rjmp    _PrintString
        .endfunc

NotSuspended:
        .asciz  " - Not Suspended!\r"
        .balign 2
;
; GetObjectID
;
; Little helper routine to get and return a byte value with
; carry set if all Ok.  e.g. entered 2 hex digits, or terminated
; with whitespace.
;
; Puts the ID in the R25:R24 register because that is what all the
; AvrX routines expect.
;
        .func   GetObjectID
GetObjectID:
        rcall   GetHexWord
        brts    goi00
        rcall   IsWhite
        brtc    goi02
goi00:
        tst     R25
        brne    goi01           ; 
        lds     Yl, ObjectId+1  ; Restore default
        lds     Yh, ObjectId
goi01:
        sts     ObjectId+1, Yl
        sts     ObjectId, Yh
        mov     R24, Yl          ; Some like Xl
        mov     R25, Yh          ; and some like Zl
goi02:
        ret
        .endfunc
;
; GetDumpLength
;
; Another helper to manage defaults with the dump length
; counter
;
        .func   GetDumpLength
GetDumpLength:
        rcall   GetHexByte      ; get a dump counter
        brts    gdl00           ; If term invalid, 
        rcall   IsWhite         ;    check for whitespace
        brtc    gdl02
gdl00:
        tst     R25             ; If no user entry
        brne    gdl01
        lds     Yl, DumpLength  ;   load up default
gdl01:
        sts     DumpLength, Yl
        mov     Xh, Yl
gdl02:
        ret
        .endfunc
;
; GetFlagsAndAddress
;
; Parse input stream for byte/word flag and address.  All are optional
; as long as the input stream is terminated with whitespace.
;
        .func   GetFlagsAndAddress
GetFlagsAndAddress:
        clt
        rcall   _GetChar
        cpi     RxByte, 's'
        brne    gbTRYEERAM
        cbr     MonFlags, 1<<EEFLAG; If "s" , clear flag
        rjmp    gbCONTINUE
gbTRYEERAM:
        cpi     RxByte, 'e'     ; if "e", set flag
        brne    gbDONE
        sbr     MonFlags, 1<<EEFLAG
gbCONTINUE:
        rcall   GetHexWord      ; Go for the address
        brtc    gbCHECKWORDFLAG
        cbr     MonFlags, 1<<WORDFLAG; Full address entered, no "w"
        rjmp    gbUPDATEADDRESS
gbCHECKWORDFLAG:
        cbr     MonFlags, 1<<WORDFLAG
        cpi     RxByte, 'w'
        brne    gbCHECKFORWHITESPACE; Only "w" allowed if not address
        sbr     MonFlags, 1<<WORDFLAG
        rcall   GetHexWord      ; Try again for address
        brts    gbUPDATEADDRESS
gbCHECKFORWHITESPACE:
        rcall   IsWhite
        brtc    gbDONE          ; Error if not whitespace
gbUPDATEADDRESS:
        mov     Zl, Yl
        mov     Zh, Yh
        ldi     Yl, lo8(Address)
        ldi     Yh, hi8(Address)
        sbrs    MonFlags, EEFLAG
        rjmp    gbUPDATE
        ldi     Yl, lo8(EEAddress)
        ldi     Yh, hi8(EEAddress)
gbUPDATE:
        tst     R25             ; Dont update address if none entered
        breq    gbDONE
        std     Y+0, Zh
        std     Y+1, Zl
gbDONE:
        ldd     Zh, Y+0
        ldd     Zl, Y+1
        ret
        .endfunc
;
;---------------------------------------------
; _DumpMemory
;
; Passed:       Z = Start Address
;               Xl = Byte/Word/Sram/EEram flag
;               R9 = number of bytes to dump
; Returns:      Nothing
; Uses:         Z, Y, Xl
;
        .func   _DumpMemory
_DumpMemory:
        sbrc    MonFlags, WORDFLAG
        cbr     Zl, 0x01        ; force even address to start.
        inc     R19
        sbrc    MonFlags, WORDFLAG
        cbr     R19, 0x01       ; Force even number of bytes to dump
m10c:
        rcall   PutCR
        mov     Yh, Zh          ; Spit out address
        mov     Yl, Zl
        rcall   PutHexWord
m10d:
        rcall   PutSpace
        rcall   GetByteWord
        sbrc    MonFlags, WORDFLAG
        rjmp    m11
        rcall   PutHexByte
        adiw    Zl, 1
        dec     R19
        rjmp    m12
m11:
        rcall   PutHexWord
        adiw    Zl, 2
        subi    R19, 2
m12:
        breq    m10exit
m10e:
        mov     Yl, Zl
        cbr     Yl, 0xF0
        breq    m10c            ; New line every 16 bytes.
        rjmp    m10d
m10exit:
        ret
        .endfunc
;
; _ModifyMemory
;
; PASSED:       Z = address, Xl = Flags
;
        .func   _ModifyMemory
_ModifyMemory:
        rcall   PutCR
        mov     Yl, Zl
        mov     Yh, Zh
        rcall   PutHexWord
        rcall   PutSpace
_mmNEXT:
        rcall   GetByteWord
        sbrc    MonFlags, WORDFLAG
        rcall   PutHexWord
        sbrs    MonFlags, WORDFLAG
        rcall   PutHexByte
_mmGETDATA:
        ldi     TxByte, '-'
        rcall   PutChar
        sbrs    MonFlags, WORDFLAG
        rcall   GetHexByte
        sbrc    MonFlags, WORDFLAG
        rcall   GetHexWord
        brtc    _mmNODATA
_mmWRITE:
        rcall   PutByteWord
        rcall   PutSpace
_mmGOTONEXT:
        adiw    Zl, 1
        sbrc    MonFlags, WORDFLAG
        adiw    Zl, 1
        mov     Yl, Zl
        cbr     Yl, 0xF0
        breq    _ModifyMemory   ; Start a new line every 16 bytes
        rjmp    _mmNEXT
_mmNODATA:
        rcall   IsWhite
        brtc    _mmDONE         ; If not white, then error
        tst     R25
        brne    _mmWRITE        ; If some data entered, then write
        cpi     RxByte, 0x0D    ; if CR then exit
        brne    _mmGOTONEXT     ; otherwize, just go to next byte
_mmDONE:
        ret
        .endfunc
;----------------------------------------------
; _PrintContext
;
; Prints a pretty picture of the context of a process.
;
; PASSED:       R25:R24 = PID
; RETURNS:      
; USES:
;
ctdt0:  .asciz  "\rPID Flg Pri Frame SR\r"
ctdt1:  .asciz  "R0  "
ctdt2:  .asciz  "R16 "
ctdt3:  .asciz  "\r     PC  "
        .balign 2
        
        .func   _PrintContext
_PrintContext:
        ldi     Zh, hi8(ctdt0)
        ldi     Zl, lo8(ctdt0)
        rcall   _PrintString
        
        mov     Zl, R24
        mov     Zh, R25         // R24:R25 get clobbered...
        
        mov     Yl, Zl
        mov     Yh, Zh
        rcall   PutHexWord      ; Print PID
        rcall   PutSpace

        ldd     Yl, Z+PidState
        rcall   PutHexByte
        rcall   PutSpace
        ldd     Yl, Z+PidPriority
        rcall   PutHexByte
        rcall   PutSpace
        rcall   PutSpace

        ldd     Yl, Z+PidSP+1   ; Z -> Frame
        ldd     Yh, Z+PidSP
        mov     Xl, Yl
        mov     Xh, Yh
        rcall   PutHexWord      ; Print Frame Pointer
        rcall   PutSpace
        rcall   PutSpace

        adiw    Xl, 1           ; Remember Stack are post dec!

        ld      Yl, X+          ; SREG
        rcall   PutHexByte
        ;
        ; Print R0-R15
        ;
Loop:
        ldi     Zh, hi8(ctdt1)
        ldi     Zl, lo8(ctdt1)
        rcall   Print16Reg
        ldi     Zh, hi8(ctdt2)
        ldi     Zl, lo8(ctdt2)
        rcall   Print16Reg
        ldi     Zh, hi8(ctdt3)
        ldi     Zl, lo8(ctdt3)
        rcall   _PrintString
        ld      Yh, X+
        ld      Yl, X
        rcall   PutHexWord
        ret

Print16Reg:
        rcall   PutCR
        mov     Yl, Xl
        mov     Yh, Xh
        rcall   PutHexWord
        rcall   PutSpace
        rcall   _PrintString
        ldi     Zl, 16
1:
        ld      Yl, X+
        rcall   PutHexByte
        rcall   PutSpace
        dec     Zl
        brne    1b
        ret
        .endfunc
;-----------------------------------------------
; GetByteWord
;
; Get a byte/word to Yl/Y from address Z in either Sram or EEram space
;
; PASSED:       Z = address
; RETURNS:      Y = data
; USES:
;
        .func   GetByteWord
GetByteWord:
        sbrc    MonFlags, EEFLAG
        rjmp    gbwEEPROM

        ldd     Yl, Z+0
        sbrs    MonFlags, WORDFLAG
        ret

        mov     Yh, Yl
        ldd     Yl, Z+1
        ret
gbwEEPROM:                      // With new calls must preserve Z
        mov     R15, Zh
        mov     R14, Zl
        mov     R25, R15
        mov     R24, R14
        rcall   AvrXReadEEProm
        mov     Yl, R24
        sbrs    MonFlags, WORDFLAG
        rjmp    gwbEXIT
        mov     Yh, Yl
        mov     R25, R15
        mov     R24, R14
        adiw    R24, 1
        rcall   AvrXReadEEProm
        mov     Yl, R24
gwbEXIT:
        mov     Zl, R14         // Restore Z
        mov     Zh, R15
        ret
        .endfunc
;
; PutByteWord
;
; Helper routine to write Yl to Z in either Sram or EEram space
;
        .func   PutByteWord
PutByteWord:
        sbrc    MonFlags, EEFLAG
        rjmp    pbwEEPROM

        std     Z+0, Yl
        sbrs    MonFlags, WORDFLAG
        ret

        std     Z+0, Yh
        std     Z+1, Yl
        ret
pbwEEPROM:
        mov     R15, Zh
        mov     R14, Zl
        mov     R25, R15
        mov     R24, R14
        mov     R23, Yl
        rcall   AvrXWriteEEProm
        sbrs    MonFlags, WORDFLAG
        rjmp    pbwEXIT

        mov     R25, R15
        mov     R24, R14
        adiw    R24, 1
        mov     R23, Yl
        rcall   AvrXWriteEEProm
        mov     R25, R15
        mov     R24, R14
        mov     R23, Yh
        rcall   AvrXWriteEEProm
pbwEXIT:
        mov     Zl, R14         // Restore Z
        mov     Zh, R15
        ret
        .endfunc
        
